@@ -36,6 +36,7 @@
 #include <glib/gi18n-lib.h>
 #include <nm-setting-connection.h>
 #include <nm-setting-8021x.h>
+#include <nm-utils.h>
 
 #include "auth-helpers.h"
 #include "nm-openvpn.h"
@@ -46,6 +47,8 @@
 #define PW_TYPE_ASK    1
 #define PW_TYPE_UNUSED 2
 
+#define BLOCK_HANDLER_ID "block-handler-id"
+
 static void
 show_password (GtkToggleButton *togglebutton, GtkEntry *password_entry)
 {
@@ -81,11 +84,25 @@ setup_secret_widget (GtkBuilder *builder,
 	return widget;
 }
 
+typedef struct {
+	GtkWidget *widget1;
+	GtkWidget *widget2;
+} TlsChooserSignalData;
+
 static void
-tls_cert_changed_cb (GtkWidget *widget, GtkWidget *next_widget)
+tls_chooser_signal_data_destroy (gpointer data, GClosure *closure)
 {
-	GtkFileChooser *this, *next;
-	char *fname, *next_fname;
+	g_slice_free (TlsChooserSignalData, data);
+}
+
+static void
+tls_cert_changed_cb (GtkWidget *widget, gpointer data)
+{
+	GtkWidget *other_widgets[2] = { ((TlsChooserSignalData *) data)->widget1,
+	                                ((TlsChooserSignalData *) data)->widget2 };
+	GtkFileChooser *this, *others[2];
+	char *fname, *other_fnames[2];
+	int i;
 
 	/* If the just-changed file chooser is a PKCS#12 file, then all of the
 	 * TLS filechoosers have to be PKCS#12.  But if it just changed to something
@@ -97,29 +114,42 @@ tls_cert_changed_cb (GtkWidget *widget, GtkWidget *next_widget)
 	 */
 
 	this = GTK_FILE_CHOOSER (widget);
-	next = GTK_FILE_CHOOSER (next_widget);
+	others[0] = GTK_FILE_CHOOSER (other_widgets[0]);
+	others[1] = GTK_FILE_CHOOSER (other_widgets[1]);
 
 	fname = gtk_file_chooser_get_filename (this);
+	other_fnames[0] = gtk_file_chooser_get_filename (others[0]);
+	other_fnames[1] = gtk_file_chooser_get_filename (others[1]);
+
 	if (is_pkcs12 (fname)) {
 		/* Make sure all choosers have this PKCS#12 file */
-		next_fname = gtk_file_chooser_get_filename (next);
-		if (!next_fname || strcmp (fname, next_fname)) {
-			/* Next chooser was different, make it the same as the first */
-			gtk_file_chooser_set_filename (next, fname);
+		for (i = 0; i < 2; i++) {
+			if (!other_fnames[i] || strcmp (fname, other_fnames[i])) {
+				/* Next chooser was different, make it the same as the first */
+				gulong id = GPOINTER_TO_SIZE (g_object_get_data (G_OBJECT (other_widgets[i]),
+				                                                 BLOCK_HANDLER_ID));
+				g_signal_handler_block (other_widgets[i], id);
+				gtk_file_chooser_set_filename (others[i], fname);
+				g_signal_handler_unblock (other_widgets[i], id);
+			}
+		}
+	} else {
+		/* Just-chosen file isn't PKCS#12 or no file was chosen, so clear out other
+		 * file selectors that have PKCS#12 files in them.
+		 */
+		for (i = 0; i < 2; i++) {
+			if (is_pkcs12 (other_fnames[i])) {
+				gulong id = GPOINTER_TO_SIZE (g_object_get_data (G_OBJECT (other_widgets[i]),
+				                                                 BLOCK_HANDLER_ID));
+				g_signal_handler_block (other_widgets[i], id);
+				gtk_file_chooser_unselect_all (others[i]);
+				g_signal_handler_unblock (other_widgets[i], id);
+			}
 		}
-		g_free (fname);
-		g_free (next_fname);
-		return;
 	}
 	g_free (fname);
-
-	/* Just-chosen file isn't PKCS#12 or no file was chosen, so clear out other
-	 * file selectors that have PKCS#12 files in them.
-	 */
-	next_fname = gtk_file_chooser_get_filename (next);
-	if (is_pkcs12 (next_fname))
-		gtk_file_chooser_set_filename (next, NULL);
-	g_free (next_fname);
+	g_free (other_fnames[0]);
+	g_free (other_fnames[1]);
 }
 
 static void
@@ -135,6 +165,8 @@ tls_setup (GtkBuilder *builder,
 	const char *value;
 	char *tmp;
 	GtkFileFilter *filter;
+	TlsChooserSignalData *ca_chooser_data, *cert_data, *key_data;
+	gulong id1, id2, id3;
 
 	tmp = g_strdup_printf ("%s_user_cert_chooser", prefix);
 	cert = GTK_WIDGET (gtk_builder_get_object (builder, tmp));
@@ -172,10 +204,28 @@ tls_setup (GtkBuilder *builder,
 			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (key), value);
 	}
 
+	ca_chooser_data = g_slice_new0 (TlsChooserSignalData);
+	ca_chooser_data->widget1 = cert;
+	ca_chooser_data->widget2 = key;
+	cert_data = g_slice_new0 (TlsChooserSignalData);
+	cert_data->widget1 = ca_chooser;
+	cert_data->widget2 = key;
+	key_data = g_slice_new0 (TlsChooserSignalData);
+	key_data->widget1 = ca_chooser;
+	key_data->widget2 = cert;
+
 	/* Link choosers to the PKCS#12 changer callback */
-	g_signal_connect (ca_chooser, "selection-changed", G_CALLBACK (tls_cert_changed_cb), cert);
-	g_signal_connect (cert, "selection-changed", G_CALLBACK (tls_cert_changed_cb), key);
-	g_signal_connect (key, "selection-changed", G_CALLBACK (tls_cert_changed_cb), ca_chooser);
+	id1 = g_signal_connect_data (ca_chooser, "selection-changed", G_CALLBACK (tls_cert_changed_cb),
+	                             ca_chooser_data, tls_chooser_signal_data_destroy, 0);
+	id2 = g_signal_connect_data (cert, "selection-changed", G_CALLBACK (tls_cert_changed_cb),
+	                             cert_data, tls_chooser_signal_data_destroy, 0);
+	id3 = g_signal_connect_data (key, "selection-changed", G_CALLBACK (tls_cert_changed_cb),
+	                             key_data, tls_chooser_signal_data_destroy, 0);
+
+	/* Store handler id to be able to block the signal in tls_cert_changed_cb() */
+	g_object_set_data (G_OBJECT (ca_chooser), BLOCK_HANDLER_ID, GSIZE_TO_POINTER (id1));
+	g_object_set_data (G_OBJECT (cert), BLOCK_HANDLER_ID, GSIZE_TO_POINTER (id2));
+	g_object_set_data (G_OBJECT (key), BLOCK_HANDLER_ID, GSIZE_TO_POINTER (id3));
 
 	/* Fill in the private key password */
 	tmp = g_strdup_printf ("%s_private_key_password_entry", prefix);
@@ -934,6 +984,8 @@ static const char *advanced_keys[] = {
 	NM_OPENVPN_KEY_TUNNEL_MTU,
 	NM_OPENVPN_KEY_FRAGMENT_SIZE,
 	NM_OPENVPN_KEY_TAP_DEV,
+	NM_OPENVPN_KEY_DEV,
+	NM_OPENVPN_KEY_DEV_TYPE,
 	NM_OPENVPN_KEY_PROTO_TCP,
 	NM_OPENVPN_KEY_PROXY_TYPE,
 	NM_OPENVPN_KEY_PROXY_SERVER,
@@ -941,12 +993,14 @@ static const char *advanced_keys[] = {
 	NM_OPENVPN_KEY_PROXY_RETRY,
 	NM_OPENVPN_KEY_HTTP_PROXY_USERNAME,
 	NM_OPENVPN_KEY_CIPHER,
+	NM_OPENVPN_KEY_KEYSIZE,
 	NM_OPENVPN_KEY_AUTH,
 	NM_OPENVPN_KEY_TA_DIR,
 	NM_OPENVPN_KEY_TA,
 	NM_OPENVPN_KEY_RENEG_SECONDS,
 	NM_OPENVPN_KEY_TLS_REMOTE,
 	NM_OPENVPN_KEY_REMOTE_RANDOM,
+	NM_OPENVPN_KEY_REMOTE_CERT_TLS,
 	NULL
 };
 
@@ -989,42 +1043,10 @@ advanced_dialog_new_hash_from_connection (NMConnection *connection,
 }
 
 static void
-port_toggled_cb (GtkWidget *check, gpointer user_data)
+checkbox_toggled_update_widget_cb (GtkWidget *check, gpointer user_data)
 {
-	GtkBuilder *builder = (GtkBuilder *) user_data;
-	GtkWidget *widget;
+	GtkWidget *widget = (GtkWidget*) user_data;
 
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "port_spinbutton"));
-	gtk_widget_set_sensitive (widget, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check)));
-}
-
-static void
-tunmtu_toggled_cb (GtkWidget *check, gpointer user_data)
-{
-	GtkBuilder *builder = (GtkBuilder *) user_data;
-	GtkWidget *widget;
-
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "tunmtu_spinbutton"));
-	gtk_widget_set_sensitive (widget, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check)));
-}
-
-static void
-fragment_toggled_cb (GtkWidget *check, gpointer user_data)
-{
-	GtkBuilder *builder = (GtkBuilder *) user_data;
-	GtkWidget *widget;
-
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "fragment_spinbutton"));
-	gtk_widget_set_sensitive (widget, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check)));
-}
-
-static void
-reneg_toggled_cb (GtkWidget *check, gpointer user_data)
-{
-	GtkBuilder *builder = (GtkBuilder *) user_data;
-	GtkWidget *widget;
-
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "reneg_spinbutton"));
 	gtk_widget_set_sensitive (widget, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check)));
 }
 
@@ -1204,6 +1226,53 @@ populate_hmacauth_combo (GtkComboBox *box, const char *hmacauth)
 	g_object_unref (store);
 }
 
+static void
+remote_tls_cert_toggled_cb (GtkWidget *widget, gpointer user_data)
+{
+	GtkBuilder *builder = (GtkBuilder *) user_data;
+	gboolean use_remote_cert_tls = FALSE;
+
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "remote_cert_tls_checkbutton"));
+	use_remote_cert_tls = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
+
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "remote_cert_tls_label"));
+	gtk_widget_set_sensitive (widget, use_remote_cert_tls);
+
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "remote_cert_tls_combo"));
+	gtk_widget_set_sensitive (widget, use_remote_cert_tls);
+}
+
+#define REMOTE_CERT_COL_NAME 0
+#define REMOTE_CERT_COL_VALUE 1
+
+static void
+populate_remote_cert_tls_combo (GtkComboBox *box, const char *remote_cert)
+{
+	GtkListStore *store;
+	GtkTreeIter iter;
+
+	store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
+	gtk_combo_box_set_model (box, GTK_TREE_MODEL (store));
+
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter,
+	                    REMOTE_CERT_COL_NAME, _("Server"),
+	                    REMOTE_CERT_COL_VALUE, NM_OPENVPN_REM_CERT_TLS_SERVER,
+	                    -1);
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter,
+	                    REMOTE_CERT_COL_NAME, _("Client"),
+	                    REMOTE_CERT_COL_VALUE, NM_OPENVPN_REM_CERT_TLS_CLIENT,
+	                    -1);
+
+	if (g_strcmp0 (remote_cert, NM_OPENVPN_REM_CERT_TLS_CLIENT) == 0)
+		gtk_combo_box_set_active (box, 1);
+	else
+		gtk_combo_box_set_active (box, 0);
+
+	g_object_unref (store);
+}
+
 static void
 tls_auth_toggled_cb (GtkWidget *widget, gpointer user_data)
 {
@@ -1229,6 +1298,9 @@ tls_auth_toggled_cb (GtkWidget *widget, gpointer user_data)
 #define PROXY_TYPE_HTTP  1
 #define PROXY_TYPE_SOCKS 2
 
+#define DEVICE_TYPE_IDX_TUN     0
+#define DEVICE_TYPE_IDX_TAP     1
+
 static void
 proxy_type_changed (GtkComboBox *combo, gpointer user_data)
 {
@@ -1288,6 +1360,86 @@ show_proxy_password_toggled_cb (GtkCheckButton *button, gpointer user_data)
 	gtk_entry_set_visibility (GTK_ENTRY (widget), visible);
 }
 
+static void
+device_name_filter_cb (GtkEntry *entry,
+                       const gchar *text,
+                       gint length,
+                       gint *position,
+                       void *user_data)
+{
+	int i, count = 0;
+	gchar *result = g_new (gchar, length + 1);
+	GtkEditable *editable = GTK_EDITABLE (entry);
+
+	for (i = 0; i < length; i++) {
+		if (text[i] == '/' || g_ascii_isspace (text[i]))
+			continue;
+		result[count++] = text[i];
+	}
+	result[count] = 0;
+
+	if (count > 0) {
+		g_signal_handlers_block_by_func (G_OBJECT (editable),
+		                                 G_CALLBACK (device_name_filter_cb),
+		                                 user_data);
+		gtk_editable_insert_text (editable, result, count, position);
+		g_signal_handlers_unblock_by_func (G_OBJECT (editable),
+		                                   G_CALLBACK (device_name_filter_cb),
+		                                   user_data);
+	}
+	g_signal_stop_emission_by_name (G_OBJECT (editable), "insert-text");
+
+	g_free (result);
+}
+
+static gboolean
+device_name_changed_cb (GtkEntry *entry,
+                        gpointer user_data)
+{
+	GtkEditable *editable = GTK_EDITABLE (entry);
+	GtkWidget *ok_button = user_data;
+	gboolean entry_sensitive;
+	char *entry_text;
+	GdkRGBA rgba;
+
+	entry_sensitive = gtk_widget_get_sensitive (GTK_WIDGET (entry));
+	entry_text = gtk_editable_get_chars (editable, 0, -1);
+
+	/* Change cell's background to red if the value is invalid */
+	if (entry_sensitive && entry_text[0] != '\0' && !nm_utils_iface_valid_name (entry_text)) {
+		gdk_rgba_parse (&rgba, "red");
+		gtk_widget_override_background_color (GTK_WIDGET (editable), GTK_STATE_NORMAL, &rgba);
+		gtk_widget_set_sensitive (ok_button, FALSE);
+	} else {
+		gtk_widget_override_background_color (GTK_WIDGET (editable), GTK_STATE_NORMAL, NULL);
+		gtk_widget_set_sensitive (ok_button, TRUE);
+	}
+
+	g_free (entry_text);
+	return FALSE;
+}
+
+static void
+dev_checkbox_toggled_cb (GtkWidget *check, gpointer user_data)
+{
+	GtkBuilder *builder = (GtkBuilder *) user_data;
+	GtkWidget *combo, *entry, *ok_button;
+
+	combo = GTK_WIDGET (gtk_builder_get_object (builder, "dev_type_combo"));
+	entry = GTK_WIDGET (gtk_builder_get_object (builder, "dev_entry"));
+	ok_button = GTK_WIDGET (gtk_builder_get_object (builder, "ok_button"));
+
+	/* Set values to default ones */
+	if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check))) {
+		gtk_entry_set_text (GTK_ENTRY (entry), "");
+		gtk_combo_box_set_active (GTK_COMBO_BOX (combo), DEVICE_TYPE_IDX_TUN);
+	}
+
+	checkbox_toggled_update_widget_cb (check, combo);
+	checkbox_toggled_update_widget_cb (check, entry);
+	device_name_changed_cb (GTK_ENTRY (entry), ok_button);
+}
+
 #define TA_DIR_COL_NAME 0
 #define TA_DIR_COL_NUM 1
 
@@ -1297,8 +1449,9 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	GtkBuilder *builder;
 	GtkWidget *dialog = NULL;
 	char *ui_file = NULL;
-	GtkWidget *widget, *combo;
+	GtkWidget *widget, *combo, *spin, *entry, *ok_button;
 	const char *value, *value2;
+	const char *dev, *dev_type, *tap_dev;
 	GtkListStore *store;
 	GtkTreeIter iter;
 	guint32 active = PROXY_TYPE_NONE;
@@ -1329,8 +1482,11 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	                        builder, (GDestroyNotify) g_object_unref);
 	g_object_set_data (G_OBJECT (dialog), "connection-type", GINT_TO_POINTER (contype));
 
+	ok_button = GTK_WIDGET (gtk_builder_get_object (builder, "ok_button"));
+
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "reneg_checkbutton"));
-	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (reneg_toggled_cb), builder);
+	spin = GTK_WIDGET (gtk_builder_get_object (builder, "reneg_spinbutton"));
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (checkbox_toggled_update_widget_cb), spin);
 
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_RENEG_SECONDS);
 	if (value && strlen (value)) {
@@ -1427,7 +1583,8 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (show_proxy_password_toggled_cb), builder);
 
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "port_checkbutton"));
-	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (port_toggled_cb), builder);
+	spin = GTK_WIDGET (gtk_builder_get_object (builder, "port_spinbutton"));
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (checkbox_toggled_update_widget_cb), spin);
 
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_PORT);
 	if (value && strlen (value)) {
@@ -1453,7 +1610,8 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "tunmtu_checkbutton"));
 	g_assert (widget);
-	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (tunmtu_toggled_cb), builder);
+	spin = GTK_WIDGET (gtk_builder_get_object (builder, "tunmtu_spinbutton"));
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (checkbox_toggled_update_widget_cb), spin);
 
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_TUNNEL_MTU);
 	if (value && strlen (value)) {
@@ -1477,7 +1635,8 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	}
 
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "fragment_checkbutton"));
-	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (fragment_toggled_cb), builder);
+	spin = GTK_WIDGET (gtk_builder_get_object (builder, "fragment_spinbutton"));
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (checkbox_toggled_update_widget_cb), spin);
 
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_FRAGMENT_SIZE);
 	if (value && strlen (value)) {
@@ -1524,11 +1683,38 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
 	}
 
-	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_TAP_DEV);
-	if (value && !strcmp (value, "yes")) {
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "tap_checkbutton"));
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-	}
+	/* Populate device-related widgets */
+	dev =      g_hash_table_lookup (hash, NM_OPENVPN_KEY_DEV);
+	dev_type = g_hash_table_lookup (hash, NM_OPENVPN_KEY_DEV_TYPE);
+	tap_dev =  g_hash_table_lookup (hash, NM_OPENVPN_KEY_TAP_DEV);
+
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "dev_checkbutton"));
+	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), (dev && *dev) || dev_type || tap_dev);
+	dev_checkbox_toggled_cb (widget, builder);
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (dev_checkbox_toggled_cb), builder);
+	combo = GTK_WIDGET (gtk_builder_get_object (builder, "dev_type_combo"));
+	active = DEVICE_TYPE_IDX_TUN;
+	if (   !g_strcmp0 (dev_type, "tap")
+	    || (!dev_type && dev && g_str_has_prefix (dev, "tap"))
+	    || (!dev_type && !g_strcmp0 (tap_dev, "yes")))
+		active = DEVICE_TYPE_IDX_TAP;
+
+	store = gtk_list_store_new (1, G_TYPE_STRING);
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter, 0, _("TUN"), -1);
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter, 0, _("TAP"), -1);
+	gtk_combo_box_set_model (GTK_COMBO_BOX (combo), GTK_TREE_MODEL (store));
+	g_object_unref (store);
+	gtk_combo_box_set_active (GTK_COMBO_BOX (combo), active);
+
+	entry = GTK_WIDGET (gtk_builder_get_object (builder, "dev_entry"));
+	gtk_entry_set_max_length (GTK_ENTRY (entry), 15);  /* interface name is max 15 chars */
+	gtk_entry_set_placeholder_text (GTK_ENTRY (entry), _("(automatic)"));
+	g_signal_connect (G_OBJECT (entry), "insert-text", G_CALLBACK (device_name_filter_cb), NULL);
+	g_signal_connect (G_OBJECT (entry), "changed", G_CALLBACK (device_name_changed_cb), ok_button);
+	if (dev && dev[0] != '\0')
+		gtk_entry_set_text (GTK_ENTRY (entry), dev);
 
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_REMOTE_RANDOM);
 	if (value && !strcmp (value, "yes")) {
@@ -1540,6 +1726,32 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_CIPHER);
 	populate_cipher_combo (GTK_COMBO_BOX (widget), value);
 
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "keysize_checkbutton"));
+	g_assert (widget);
+	spin = GTK_WIDGET (gtk_builder_get_object (builder, "keysize_spinbutton"));
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (checkbox_toggled_update_widget_cb), spin);
+
+	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_KEYSIZE);
+	if (value && strlen (value)) {
+		long int tmp;
+
+		errno = 0;
+		tmp = strtol (value, NULL, 10);
+		if (errno == 0 && tmp > 0 && tmp < 65536) {
+			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
+
+			widget = GTK_WIDGET (gtk_builder_get_object (builder, "keysize_spinbutton"));
+			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gdouble) tmp);
+			gtk_widget_set_sensitive (widget, TRUE);
+		}
+	} else {
+		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), FALSE);
+
+		widget = GTK_WIDGET (gtk_builder_get_object (builder, "keysize_spinbutton"));
+		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), 128.0);
+		gtk_widget_set_sensitive (widget, FALSE);
+	}
+
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "hmacauth_combo"));
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_AUTH);
 	populate_hmacauth_combo (GTK_COMBO_BOX (widget), value);
@@ -1550,6 +1762,17 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 		gtk_entry_set_text (GTK_ENTRY(widget), value);
 	}
 
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "remote_cert_tls_checkbutton"));
+	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_REMOTE_CERT_TLS);
+	if (value && strlen (value))
+		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (remote_tls_cert_toggled_cb), builder);
+	remote_tls_cert_toggled_cb (widget, builder);
+
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "remote_cert_tls_combo"));
+	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_REMOTE_CERT_TLS);
+	populate_remote_cert_tls_combo (GTK_COMBO_BOX (widget), value);
+
 	if (   !strcmp (contype, NM_OPENVPN_CONTYPE_TLS)
 	    || !strcmp (contype, NM_OPENVPN_CONTYPE_PASSWORD_TLS)
 	    || !strcmp (contype, NM_OPENVPN_CONTYPE_PASSWORD)) {
@@ -1731,9 +1954,21 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog, GError **error)
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
 		g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_PROTO_TCP), g_strdup ("yes"));
 
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "tap_checkbutton"));
-	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
-		g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_TAP_DEV), g_strdup ("yes"));
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "dev_checkbutton"));
+	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
+		int device_type;
+
+		widget = GTK_WIDGET (gtk_builder_get_object (builder, "dev_type_combo"));
+		device_type = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
+		g_hash_table_insert (hash,
+		                     g_strdup (NM_OPENVPN_KEY_DEV_TYPE),
+		                     g_strdup (device_type == DEVICE_TYPE_IDX_TUN ? "tun" : "tap"));
+
+		widget = GTK_WIDGET (gtk_builder_get_object (builder, "dev_entry"));
+		value = gtk_entry_get_text (GTK_ENTRY (widget));
+		if (value && value[0] != '\0')
+			g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_DEV), g_strdup (value));
+	}
 
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "remote_random_checkbutton"));
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
@@ -1753,6 +1988,15 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog, GError **error)
 		}
 	}
 
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "keysize_checkbutton"));
+	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
+		int keysize_val;
+
+		widget = GTK_WIDGET (gtk_builder_get_object (builder, "keysize_spinbutton"));
+		keysize_val = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (widget));
+		g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_KEYSIZE), g_strdup_printf ("%d", keysize_val));
+	}
+
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "hmacauth_combo"));
 	model = gtk_combo_box_get_model (GTK_COMBO_BOX (widget));
 	if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (widget), &iter)) {
@@ -1777,6 +2021,21 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog, GError **error)
 		if (value && strlen (value))
 			g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_TLS_REMOTE), g_strdup (value));
 
+		widget = GTK_WIDGET (gtk_builder_get_object (builder, "remote_cert_tls_checkbutton"));
+		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
+			widget = GTK_WIDGET (gtk_builder_get_object (builder, "remote_cert_tls_combo"));
+			model = gtk_combo_box_get_model (GTK_COMBO_BOX (widget));
+			if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (widget), &iter)) {
+				char *remote_cert = NULL;
+
+				gtk_tree_model_get (model, &iter, REMOTE_CERT_COL_VALUE, &remote_cert, -1);
+				if (remote_cert)
+					g_hash_table_insert (hash,
+					                     g_strdup (NM_OPENVPN_KEY_REMOTE_CERT_TLS),
+					                     remote_cert);
+			}
+		}
+
 		widget = GTK_WIDGET (gtk_builder_get_object (builder, "tls_auth_checkbutton"));
 		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
 			char *filename;
